<?php
/*
Plugin Name: 5sec Google Authenticator - Two Step Login Protection
Plugin URI: http://5sec-google-authenticator.webfactoryltd.com/
Description: Add bank grade security to your site using one time passwords. Additional security is provided by brute-force attack protection and auto log out feature.
Author: پرشین اسکریپت
Version: 1.20
Author URI: http://www.persianscript.ir
*/


if (!function_exists('add_action')) {
  die('Please don\'t open this file directly!');
}

define('WF_GA_VER', '1.15');
define('WF_GA_OPTIONS', 'wf_ga');
define('WF_GA_TABLE_LOG', 'wf_ga_log');
define('WF_GA_TABLE_BAN', 'wf_ga_ban');


require_once 'wf_ga_base32.php';


// drop in function because there are no filters/actions in the original one
// originally in wp-includes/pluggable.php:1190
if (!function_exists('wp_new_user_notification')) {
  function wp_new_user_notification($user_id, $plaintext_pass = '') {
    $user = get_userdata( $user_id );
    $options = get_option(WF_GA_OPTIONS);

    // The blogname option is escaped with esc_html on the way into the database in sanitize_option
    // we want to reverse this for the plain text arena of emails.
    $blogname = wp_specialchars_decode(get_option('blogname'), ENT_QUOTES);

    $message  = sprintf(__('New user registration on your site %s:'), $blogname) . "\r\n\r\n";
    $message .= sprintf(__('Username: %s'), $user->user_login) . "\r\n\r\n";
    $message .= sprintf(__('E-mail: %s'), $user->user_email) . "\r\n";

    @wp_mail(get_option('admin_email'), sprintf(__('[%s] New User Registration'), $blogname), $message);

    if ( empty($plaintext_pass) )
      return;

    if ($options['send_qr_users']) {
      $qr_img = wf_ga::get_qr_string($user_id);
      $qr_img = $qr_img['img_url'];

      $message  = sprintf(__('Username: %s'), $user->user_login) . '<br>';
      $message .= sprintf(__('Password: %s'), $plaintext_pass) . '<br>';
      $message .= 'Login URL: <a href="' . wp_login_url() . '">' . wp_login_url() . '</a><br>';
      $message .= __('Google Authenticator QR Code:', 'wf_ga') . '<br>';
      $message .= '<img alt="' . __('Google Authenticator QR Code', 'wf_ga') . '" title="' . __('Google Authenticator QR Code', 'wf_ga') . '" src="' . $qr_img . '" />' . '<br>';
      $message .= sprintf(__('If the QR image is not visible <a href="%s">open it in browser</a>.', 'wf_ga'), $qr_img);

      wp_mail($user->user_email, sprintf(__('[%s] Your username and password'), $blogname), $message, 'Content-type: text/html; charset=UTF-8');
    } else {
      $message  = sprintf(__('Username: %s'), $user->user_login) . "\r\n";
      $message .= sprintf(__('Password: %s'), $plaintext_pass) . "\r\n";
      $message .= wp_login_url() . "\r\n";

      wp_mail($user->user_email, sprintf(__('[%s] Your username and password'), $blogname), $message);
    }
  } // wp_new_user_notification
}


// main plugin class
class wf_ga {
  // init plugin
  static function init() {
    // first things first
    self::check_ban();

    if (!session_id()) {
      session_start();
    }

    if (is_admin()) {
      // this plugin requires WP v3.6 (only because of heartbeat)
      if (!version_compare(get_bloginfo('version'), '3.6',  '>=')) {
        add_action('admin_notices', array(__CLASS__, 'min_version_error_wp'));
      }

      // warn users to read the documentation
      if (get_transient('wf_ga_activation_notice')) {
        add_action('admin_notices', array(__CLASS__, 'activation_notice'));
      }

      // aditional links in plugin description
      add_filter('plugin_action_links_' . basename(dirname(__FILE__)) . '/' . basename(__FILE__),
                 array(__CLASS__, 'plugin_action_links'));
      add_filter('plugin_row_meta', array(__CLASS__, 'plugin_meta_links'), 10, 2);

      // enqueue scripts
      add_action('admin_enqueue_scripts', array(__CLASS__, 'enqueue_scripts'));

      // register ajax endpoints
      add_action('wp_ajax_wf_ga_generate_secret_key', array(__CLASS__, 'generate_secret_key'));
      add_action('wp_ajax_wf_ga_generate_secret_url', array(__CLASS__, 'generate_secret_url_ajax'));
      add_action('wp_ajax_wf_ga_clean_log_ban', array(__CLASS__, 'clean_log_ban_ajax'));
      add_action('wp_ajax_wf_ga_remove_bans', array(__CLASS__, 'remove_bans_ajax'));
      add_action('wp_ajax_wf_ga_remove_activation_notice', array(__CLASS__, 'remove_activation_notice_ajax'));
      add_action('wp_ajax_wf_ga_my_ip', array(__CLASS__, 'my_ip_ajax'));
      add_action('wp_ajax_wf_ga_send_qr_emails', array(__CLASS__, 'send_qr_emails_ajax'));

      // add options to user screen
      add_action('show_user_profile', array(__CLASS__, 'show_user_profile'), 10, 1);
      add_action('personal_options_update', array(__CLASS__, 'user_options_update'), 10, 1);
      add_action('edit_user_profile', array(__CLASS__, 'show_user_profile'), 10, 1);
      add_action('edit_user_profile_update', array(__CLASS__, 'user_options_update'), 10, 1);

      // hook into heartbeat API
      add_filter('heartbeat_received', array(__CLASS__, 'heartbeat_received'), 10, 2);

      // add settings menu
      add_action('admin_menu', array(__CLASS__, 'add_menu'));

      // check and set default settings
      self::default_settings(false);

      // settings registration
      add_action('admin_init', array(__CLASS__, 'register_settings'));
    } else {
      // hook custom authenticate function
      add_action('login_form', array(__CLASS__, 'login_form'));
      add_action('login_footer', array(__CLASS__, 'login_footer'));
      add_filter('authenticate', array(__CLASS__, 'authenticate'), 50, 3);

      // log failed login attempt
      add_action('wp_login_failed', array(__CLASS__, 'login_failed'));
    }

    // record last activity
    add_action('wp_login', array(__CLASS__, 'wp_login'));
    if (is_user_logged_in()) {
      self::update_last_activity();
    }

    // automatically add authenticator settings when user registers
    add_action('user_register',array(__CLASS__, 'user_register'));
  } // init


  // textdomain has to be loaded earlier
  static function plugins_loaded() {
    load_plugin_textdomain('wf_ga', false, basename(dirname(__FILE__)) . '/lang');
  } // plugins_loaded


  // record last activity time for auto logout
  static function update_last_activity($force = false, $user_id = 0) {
    if ($force || (!defined('DOING_AJAX') || !DOING_AJAX)) {
      if (!$user_id) {
        $user_id = get_current_user_id();
      }
      update_user_option($user_id, 'wf_ga_last_activity', time());
    }
  } // update_last_activity


  // misc functions on login
  static function wp_login() {
    // if bypass was enabled kill it
    $_SESSION['bypass_auth'] = false;

    self::update_last_activity();
  } // wp_login


  // log failed login attempt
  static function login_failed() {
    global $wpdb;
    $options = get_option(WF_GA_OPTIONS);

    // check if IP is whitelisted
    if ($options['ip_whitelist'] && $_SERVER['REMOTE_ADDR']
        && strpos($options['ip_whitelist'], $_SERVER['REMOTE_ADDR']) !== false) {
      return;
    }

    $query = $wpdb->prepare('INSERT INTO ' . $wpdb->prefix . WF_GA_TABLE_LOG .
                            ' (ip, timestamp)
                            VALUES (%s, %s)',
                            $_SERVER['REMOTE_ADDR'], current_time('mysql'));
    $wpdb->query($query);

    self::update_bans();
    self::clean_tables();
  } // login_failed


  // add new bans if necessary
  static function update_bans() {
    global $wpdb;
    $options = get_option(WF_GA_OPTIONS);

    $query = $wpdb->prepare('SELECT COUNT(ip) FROM ' . $wpdb->prefix . WF_GA_TABLE_LOG .
                            ' WHERE ip = %s AND timestamp >= DATE_SUB(%s, INTERVAL %s MINUTE)',
                            $_SERVER['REMOTE_ADDR'], current_time('mysql'), $options['max_login_attempts_time']);
    $login_attempts = $wpdb->get_var($query);

    if ($login_attempts >= $options['max_login_attempts']) {
      $banned_until = current_time('timestamp') + $options['ban_time'] * 60;
      $banned_until = date('Y-m-d H:i:m', $banned_until);
      $query = $wpdb->prepare('INSERT INTO ' . $wpdb->prefix . WF_GA_TABLE_BAN .
                            ' (ip, banned_until) VALUES (%s, %s) ' .
                            ' ON DUPLICATE KEY UPDATE banned_until = %s',
                            $_SERVER['REMOTE_ADDR'], $banned_until, $banned_until);
      $wpdb->query($query);
    } // add new ban
  } // update_bans


  // check if user's IP is banned
  static function check_ban() {
    global $wpdb;
    $options = get_option(WF_GA_OPTIONS);

    $query = $wpdb->prepare('SELECT ip FROM ' . $wpdb->prefix . WF_GA_TABLE_BAN .
                            ' WHERE ip = %s AND banned_until >= %s',
                            $_SERVER['REMOTE_ADDR'], current_time('mysql'));
    $tmp = $wpdb->get_var($query);

    if ($tmp && ($options['banned_users'] == '1' ||
                 $GLOBALS['pagenow'] == 'wp-login.php')) {
      wp_die(__('Your IP was banned because of too many failed login attempts. Contact your site administrator for details.', 'wf_ga'));
    }
  } // check_ban


  // clean ban and log tables
  static function clean_tables($force = false) {
    // clean only on 10/100 requests
    if ($force || wp_rand(0, 100) <= 10) {
      global $wpdb;
      $options = get_option(WF_GA_OPTIONS);

      $banned_until_max = current_time('timestamp') - $options['ban_time'] * 60;
      $banned_until_max = date('Y-m-d H:i:m', $banned_until_max);

      $query = $wpdb->prepare('DELETE FROM ' . $wpdb->prefix . WF_GA_TABLE_LOG .
                              ' WHERE timestamp < %s', $banned_until_max);
      $wpdb->query($query);
      $query = $wpdb->prepare('DELETE FROM ' . $wpdb->prefix . WF_GA_TABLE_BAN .
                              ' WHERE banned_until < %s', current_time('mysql'));
      $wpdb->query($query);
    }
  } // clean_tables


  // ajax call for cleaning tables
  static function clean_log_ban_ajax() {
    check_ajax_referer('ajax', 'ajax_nonce');

    self::clean_tables(true);

    die('1');
  } // clean_log_ban_ajax


  // adds current IP to the whitelist
  static function my_ip_ajax() {
    check_ajax_referer('ajax', 'ajax_nonce');

    echo json_encode($_SERVER['REMOTE_ADDR']);
    die();
  } // my_ip_ajax


  // sends emails containing QR code, no other info is sent
  static function send_qr_emails_ajax() {
    check_ajax_referer('ajax', 'ajax_nonce');

    $user_ids = @$_POST['user_ids'];
    $user_ids = trim(trim($user_ids), ',');
    $user_ids = explode(',', $user_ids);

    if (!sizeof($user_ids)) {
      echo '0';
      die();
    }

    $blogname = wp_specialchars_decode(get_option('blogname'), ENT_QUOTES);
    $cnt = 0;
    foreach ($user_ids as $user_id) {
      $user_id = (int) $user_id;
      $user = get_user_by('id', $user_id);
      $qr = self::get_qr_string($user_id);

      $message = '<b>' . $blogname . '</b> uses 5sec Google Authenticator to provide additional security to all users. Please follow the instructions below to ensure you can continue to login without any problems. If you have any questions please contact your site administrator.<br><br>';
      $message .= '<ol><li>Install Google Authenticator app for <a target="_blank" href="https://itunes.apple.com/hr/app/google-authenticator/id388497605?mt=8">iPhone</a>, <a target="_blank" href="https://play.google.com/store/apps/details?id=com.google.android.apps.authenticator2">Android</a> or BlackBerry</li>
  <li>Open the app, click "add an account" (plus sign icon) and choose "scan barcode". Point your phone\'s camera to the QR code on the bottom of this email. This QR is unique, generated only for your WP account!</li>
  <li>You\'re done! Next time you want to login enter the usual username &amp; password combination and the OTP the app generates. Please note that each OTP is valid for only 2 minutes so always use a fresh one.</li>
</ol>';
      $message .= '<br><br>QR code for user "' . $user->user_login . '" ';
      $message .= '<img src="' . $qr['img_url'] . '" /><br>';
      $message .= 'If the QR code image is not showing <a href="' . $qr['img_url'] . '">open it in browser</a>.';

      $cnt += wp_mail($user->user_email, sprintf(__('[%s] Your 5sec Google Authenticator QR Code'), $blogname), $message, 'Content-type: text/html; charset=UTF-8');
    } // foreach user

    echo "$cnt";
    die();
  } // send_qr_emails_ajax


  // truncate log and ban tables
  static function remove_bans_ajax() {
    check_ajax_referer('ajax', 'ajax_nonce');

    global $wpdb;

    $query = $wpdb->prepare('TRUNCATE TABLE ' . $wpdb->prefix . WF_GA_TABLE_LOG, null);
    $wpdb->query($query);
    $query = $wpdb->prepare('TRUNCATE TABLE ' . $wpdb->prefix . WF_GA_TABLE_BAN, null);
    $wpdb->query($query);

    die('1');
  } // remove_bans_ajax


  // dddd
  static function remove_activation_notice_ajax() {
    delete_transient('wf_ga_activation_notice');

    die('1');
  } // remove_activation_notice_ajax

  // auto-logout if inactive
  static function heartbeat_received($response, $data) {
    $options = get_option(WF_GA_OPTIONS);

    $response['wf_ga_last_activity'] = get_user_option('wf_ga_last_activity', get_current_user_id());
    if ($options['auto_logout'] > 0 &&
        (time() - (int) $response['wf_ga_last_activity']) > ($options['auto_logout'] * 60)) {
      wp_logout();
    }

    return $response;
  } // heartbeat_received


  // add links to plugin's description in plugins' table
  static function plugin_meta_links($links, $file) {
    $documentation_link = '<a target="_blank" href="' . plugin_dir_url(__FILE__) . 'documentation/' .
                          '" title="View documentation">' . __('Documentation', 'wf_ga') . '</a>';
    $support_link = '<a target="_blank" href="http://codecanyon.net/user/WebFactory#from" title="Contact Web factory">' . __('Support', 'wf_ga') . '</a>';

    if ($file == plugin_basename(__FILE__)) {
      $links[] = $documentation_link;
      $links[] = $support_link;
    }

    return $links;
  } // plugin_meta_links


  // add settings link to plugins page
  static function plugin_action_links($links) {
    $settings_link = '<a href="' . admin_url('options-general.php?page=wf-ga') . '" title="' . __('5sec Google Authenticator', 'wf_ga') . '">' . __('Configure plugin', 'wf_ga') . '</a>';
    array_unshift($links, $settings_link);

    return $links;
  } // plugin_action_links


  // check if we're on the plugin's page
  static function is_plugin_page() {
    $current_screen = get_current_screen();

    if ($current_screen->id == 'settings_page_wf-ga') {
      return true;
    } else {
      return false;
    }
  } // is_plugin_page


  // generate a new random secret key for the Google Authenticator
  static function generate_secret_key($return = false) {
    $chars = '234567ABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $secret = '';

    for ($i = 0; $i < 16; $i++) {
        $secret .= substr($chars, wp_rand(0, 31), 1);
    }

    if (defined('DOING_AJAX') && DOING_AJAX && !$return) {
      echo json_encode($secret);
      die();
    } else {
      return $secret;
    }
  } // generate_secret_key


  // generate secret login URL
  static function generate_secret_url_ajax() {
    check_ajax_referer('ajax', 'ajax_nonce');

    $out = array();

    $out['key'] = self::generate_secret_key(true);
    $out['url'] = site_url() . '/wp-login.php?bypass_auth=' . $out['key'];

    echo json_encode($out);
    die();
  } // generate_secret_url


  // verify user entered OTP
  static function verify_otp($secretkey, $otp) {
    // sanitize
    if (strlen($otp) != 6) {
      return false;
    } else {
      $otp = intval($otp);
    }

    // time drifting is +/- min
    $firstcount = -2;
    $lastcount  =  2;
    $tm = floor(time() / 30);

    $secretkey = wf_Base32::decode($secretkey);
    for ($i = $firstcount; $i <= $lastcount; $i++) {
      $time = chr(0).chr(0).chr(0).chr(0).pack('N*', $tm + $i);
      $hm = hash_hmac('SHA1', $time, $secretkey, true);
      $offset = ord(substr($hm, -1)) & 0x0F;
      $hashpart = substr($hm, $offset, 4);
      $value = unpack("N", $hashpart);
      $value = $value[1];
      $value = $value & 0x7FFFFFFF;
      $value = $value % 1000000;
      if ($value == $otp) {
        return true;
      }
    } // for

    return false;
  } // verify otp


  // validate OTP when user logins
  static function authenticate($user, $username = '', $password = '') {
    if (!$username) {
      return $user;
    }

    // skip if RPC login
    if ((defined('XMLRPC_REQUEST') || defined('APP_REQUEST')) && $options['disable_on_rpc']) {
      return $user;
    }

    // bypassing authenticator?
    if (self::bypass_auth()) {
      return $user;
    }

    // get user details
    $userstate = $user;
    $user = get_user_by('login', $username);
    $secret = get_user_option('wf_ga_secret_key', $user->ID);

    // if user doesn't have two step authentication enabled, skip it
    if(get_user_option('wf_ga_enable_two_step', $user->ID) === '0') {
      return $userstate;
    }

    // OTP entered by the user
    $field_name = 'wf_ga_otp_' . substr(session_id(), 0, 8);
    $otp = trim(@$_POST[$field_name]);

    if (self::verify_otp($secret, $otp)) {
      self::update_last_activity(true, $user->ID);

      return $userstate;
    } else {
        return new WP_Error('invalid_google_authenticator_otp', __('<strong>ERROR</strong>: The Google Authenticator one time password (OTP) is incorrect or has expired. Please contact your administrator if you continue having problems logging in.', 'wf_ga'));
    }
  } // authenticate


  // add plugin's setting menu entry
  static function add_menu() {
    add_options_page(__('5sec Google Authenticator', 'wf_ga'), __('5sec Google Authenticator', 'wf_ga'), 'manage_options', 'wf-ga', array(__CLASS__, 'settings_page'));
  } // add_menu


  // complete markup for settings page
  static function settings_page() {
    global $wpdb;
    $options = get_option(WF_GA_OPTIONS);

    $bruteforce_timeouts = array();
    $bruteforce_timeouts[] = array('val' => 2, 'label' => __('2 minutes', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 10, 'label' => __('10 minutes', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 20, 'label' => __('20 minutes', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 30, 'label' => __('30 minutes', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 60, 'label' => __('1 hour', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 120, 'label' => __('2 hours', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 1440, 'label' => __('1 day', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 2880, 'label' => __('2 days', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 10080, 'label' => __('7 days', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 43200, 'label' => __('1 month', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 525600, 'label' => __('1 year', 'wf_ga'));
    $bruteforce_timeouts[] = array('val' => 5256000, 'label' => __('10 years', 'wf_ga'));

    $ban_rules[] = array('val' => '1', 'label' => __("Can't access the whole site", 'wf_ga'));
    $ban_rules[] = array('val' => '2', 'label' => __("Can't login", 'wf_ga'));

    $ban_cnt = $wpdb->prepare('SELECT COUNT(id) FROM ' . $wpdb->prefix . WF_GA_TABLE_BAN .
                              ' WHERE banned_until >= %s',
                              current_time('mysql'));
    $ban_cnt = $wpdb->get_var($ban_cnt);

    //settings_errors();
    echo '<div class="wrap">';
    echo '<h2>' . __('5sec Google Authenticator Settings', 'wf_ga') . '</h2>';

    echo '<form action="options.php" method="post">';
    settings_fields(WF_GA_OPTIONS);

    echo '<h3 class="title">' . __('Two Step Login Settings', 'wf_ga') . '</h3>';

    echo '<table class="form-table"><tbody>';

    echo '<tr valign="top">
    <th scope="row"><label for="disable_on_rpc">' . __('Disable two step authentication on RPC login', 'wf_ga') . '</label></th>
    <td><input name="' . WF_GA_OPTIONS . '[disable_on_rpc]" id="disable_on_rpc" ' . checked('1', $options['disable_on_rpc'], false) . ' value="1" type="checkbox" />';
    echo '<p class="description">' . __('Applications such as WordPress for iOS wont work if two step authentication is enabled on RPC login so you can turn it off. If you do not use such apps leave it on. Default: checked.', 'wf_ga') . '</p></td></tr>';

    echo '<tr valign="top">
    <th scope="row"><label for="auto_logout">' . __('Auto logout after being idle for', 'wf_ga') . '</label></th>
    <td><input name="' . WF_GA_OPTIONS . '[auto_logout]" id="auto_logout" value="' . esc_attr($options['auto_logout']) . '" class="small-text" type="text" /> <label>' . __('minute(s)', 'wf_ga') . '</label>';
    echo '<p class="description">' . __('After being idle (not clicking anything) in the admin for the set amount of minutes a lightbox will pop-up asking for username, password and OTP. You wont navigate away from the page you are on. Enter zero to disable this feature. Default: 15 minutes.', 'wf_ga') . '</p></td></tr>';

    echo '<tr valign="top">
    <th scope="row"><label for="secret_login2">' . __('Secret login URL', 'wf_ga') . '</label></th>';
    echo '<td><input type="text" class="regular-text code" disabled="disabled" id="secret_login2" name="secret_login2" value="' . site_url() . '/wp-login.php?bypass_auth=' . esc_attr($options['secret_login']) . '"><input name="' . WF_GA_OPTIONS . '[secret_login]" id="secret_login" value="' . esc_attr($options['secret_login']) . '" class="" type="hidden" /> <input type="button" value="' . __('Generate new secret URL', 'wf_ga') . '" name="generate_secret_url" id="generate_secret_url" class="button button_secondary" />';
    echo '<p class="description">' . __('Do not share this URL around! It enables you to login without OTP in case you loose your phone.', 'wf_ga') . '</p></td></tr>';

    echo '<tr valign="top">
    <th scope="row"><label for="send_qr_users">' . __('Send QR code to new users', 'wf_ga') . '</label></th>
    <td><input name="' . WF_GA_OPTIONS . '[send_qr_users]" id="send_qr_users" ' . checked('1', $options['send_qr_users'], false) . ' value="1" type="checkbox" />';
    echo '<p class="description">' . __("When new users register and receive an email with their username and password they will receive their unique QR code as well. If you disable this feature you'll have to send the QR codes manually. Default: checked.", 'wf_ga') . '</p></td></tr>';

    echo '</tbody></table>';

    echo '<h3 class="title">' . __('Brute-force Protection Settings', 'wf_ga') . '</h3>';

    echo '<table class="form-table"><tbody>';

    echo '<tr valign="top">';
    echo '<th scope="row"><label for="wf_ln_options_mla">' . __('Maximum number of failed login attempts before ban', 'wf_ga') . '</label></th>';
    echo '<td><input type="text" class="small-text" name="' . WF_GA_OPTIONS . '[max_login_attempts]" id="max_login_attempts" value="' . esc_attr($options['max_login_attempts']) .'" />';
    echo ' <label>' . __('attempt(s) in', 'wf_ga') . '</label> ';
    echo '<input type="text" class="small-text" name="' . WF_GA_OPTIONS . '[max_login_attempts_time]" id="max_login_attempts_time" value="' . esc_attr($options['max_login_attempts_time']) . '" />';
    echo '</select> <label>' . __('minute(s)', 'wf_ga') . '</label>';
    echo '<p class="description">' . __('If a user from the same IP has the set amount of failed logins in the set amount of time their IP will be banned. Default: 5 attempts in 5 minutes.', 'wf_ga') . '</p></td></tr>';

    echo '<tr valign="top">';
    echo '<th scope="row"><label for="ban_time">' . __('Ban time', 'wf_ga') . '</label></th>';
    echo '<td><select name="' . WF_GA_OPTIONS . '[ban_time]" id="ban_time">';
    self::create_select_options($bruteforce_timeouts, $options['ban_time']);
    echo '</select>';
    echo '<p class="description">' . __('Amount of time an IP will be banned after violating the rule above. Default: 2 hours.', 'wf_ga') . '</p></td></tr>';

    echo '<tr valign="top">
    <th scope="row"><label for="banned_users">' . __('Banned users', 'wf_ga') . '</label></th>
    <td><select name="' . WF_GA_OPTIONS . '[banned_users]" id="banned_users">';
    self::create_select_options($ban_rules, $options['banned_users']);
    echo '</select>';
    echo '<p class="description">' . __("Completely banning users is quite unfriendly but necessary on some sites. Default: can't login.", 'wf_ga') . '</p></td></tr>';

    echo '<tr valign="top">
    <th scope="row"><label for="ip_whitelist">' . __('IP whitelist', 'wf_ga') . '</label></th>
    <td><textarea style="margin-bottom: -5px;" rows="3" name="' . WF_GA_OPTIONS . '[ip_whitelist]" id="ip_whitelist">';
    echo esc_html($options['ip_whitelist']);
    echo '</textarea> <input type="button" id="add_whitelist_ip" class="button button-secondary" value="Add my current IP to the list">';
    echo '<p class="description">' . __("List of IP addresses that are ignored by ban rules. Wildcards are not supported. Write one IP per line without leading zeros ie: 192.168.1.12.", 'wf_ga') . '</p></td></tr>';

    echo '</tbody></table>';
    echo '<p class="submit"><input type="submit" value="' . __('Save Settings', 'wf_ga') . '" class="button-primary" name="Submit" /></p>';

    echo '<h3 class="title">' . __('Miscellaneous', 'wf_ga') . '</h3>';
    echo '<table class="form-table"><tbody>';
    echo '<tr valign="top">';
    echo '<th scope="row" colspan=""><label>' . __('Clean log &amp; ban tables', 'wf_ga') . '</label></th>';
    echo '<td><input type="button" class="button button-secondary" id="clean_log_ban" value="' . __('Clean', 'wf_ga') . '">';
    echo '<p class="description">' . __('This will not affect any active bans. It just removes old records. Tables are periodically automatically cleaned but you can do it manually as well.', 'wf_ga') . '</p></td></tr>';
    echo '<tr valign="top">';
    echo '<th scope="row" colspan=""><label>' . __('Remove all active bans', 'wf_ga') . '</label></th>';
    echo '<td><input type="button" class="button button-secondary" id="remove_bans" value="' . __('Remove', 'wf_ga') . '"> ';
    echo '<label> ' . __('Current number of banned IPs', 'wf_ga') . ': <strong id="ban_cnt">' . $ban_cnt .'</strong>.</label>';
    echo '<p class="description">' . __('This will remove all active bans and reset failed login counters for all IPs.', 'wf_ga') . '</p></td></tr>';
    echo '<tr valign="top">';
    echo '<th scope="row" colspan=""><label>' . __('View documentation', 'wf_ga') . '</label></th>';
    echo '<td><a class="buton button-secondary" target="_blank" href="' . plugin_dir_url(__FILE__) . 'documentation/">' . __('View', 'wf_ga') . '</a>';
    echo '<p class="description">' . __('If you still have question after reading the documentation contact our', 'wf_ga') . ' <a target="_blank" href="http://codecanyon.net/user/WebFactory#from" title="Contact Web factory">' . __('support', 'wf_ga') . '</a>.</p></td></tr>';
    echo '</tbody></table>';

    echo '</form></div>'; // main container
  } // settings_page


  // composes QR image URL
  static function get_qr_string($user_id) {
    $out = array();

    $secret = get_user_option('wf_ga_secret_key', $user_id);
    $username = get_user_option('user_login', $user_id);
    $site_name = sanitize_user(str_replace(' ', '_', substr(get_bloginfo('name'), 0, 30)), true);

    $out['description'] = $site_name . ':' . $username;
    $out['otpauth'] = 'otpauth://totp/' . $out['description'] . '?secret=' . $secret . '&issuer=' . $site_name;
    $out['img_url'] = 'https://chart.googleapis.com/chart?cht=qr&amp;chs=200x200&amp;chld=H|0&amp;chl=' . urlencode($out['otpauth']);

    return $out;
  } // get_qr_string


  // adds authenticator related fields to users' profile page
  static function show_user_profile($user) {
    $secret = get_user_option('wf_ga_secret_key', $user->ID);
    $enable_two_step = get_user_option('wf_ga_enable_two_step', $user->ID);
    $qr_img = self::get_qr_string($user->ID);
    $qr_img = $qr_img['img_url'];

    if ($enable_two_step === false) {
      $enable_two_step = '1';
    }

    echo '<h3>' . __('5sec Google Authenticator', 'wf_ga') . '</h3>';
    echo '<table class="form-table">';

    if (current_user_can('edit_users')) {
      echo '<tr><th><label for="wf_ga_enable_two_step">' . __('Enable Two Step Authentication', 'wf_ga') . '</label></th>';
      echo '<td><input name="wf_ga_enable_two_step" id="wf_ga_enable_two_step" ' . checked('1', $enable_two_step, false) . ' value="1" type="checkbox" />';
      echo '<span class="description"> ' . __('Only admins can enable/disable two step authentication on a per-user level. Please note that disabling two step authentication even for one user will significantly lower your site\'s security!', 'wf_ga') . '</span></td></tr>';
    }

    echo '<tr><th><label for="wf_ga_secret_key">' . __('Secret Key', 'wf_ga') . '</label></th>';
    echo '<td><input type="hidden" name="wf_ga_secret_key" id="wf_ga_secret_key" value="' . esc_attr($secret) . '">';
    echo '<input type="text" name="wf_ga_secret_key2" id="wf_ga_secret_key2" value="';
    echo esc_attr($secret) . '" class="regular-text" disabled="disabled" /> <input type="button" class="button button-secondary" value="' . __('Generate new Secret Key', 'wf_ga') . '" id="wf_ga_generate_secret_key" /><br>
    <span class="description">' . __('If you generate a new secret key you will have to add a new entry (scan the new QR code) in your mobile authenticator app. Old one will not work. Do not change the key unless you are having problems loggin in or the key has been compromised.', 'wf_ga') . '</span></td></tr>';
    echo '<tr><th><label for="">' . __('QR Code', 'wf_ga') . '</label></th>';
    echo '<td><span class="description">' . __('Scan the QR code with your mobile authenticator app to add this account to the list of your applications/accounts.', 'wf_ga') . '</span><br /><img id="wf_ga_qr" src="' . $qr_img . '" alt="' . __('User specific authenticator QR code', 'wf_ga') . '" title="' . __('User specific authenticator QR code', 'wf_ga') . '" /> </td></tr>';
    echo '</table>';
  } // show_user_profile


  // updates authenticator related fields when user's profile is updated
  static function user_options_update($user_id) {
    $secret_key = trim($_POST['wf_ga_secret_key']);
    if (strlen($secret_key) != 16) {
      $secret_key = self::generate_secret_key();
    }

    if (current_user_can('edit_users')) {
      if (isset($_POST['wf_ga_enable_two_step'])) {
        update_user_option($user_id, 'wf_ga_enable_two_step', '1');
      } else {
        update_user_option($user_id, 'wf_ga_enable_two_step', '0');
      }
    }

    update_user_option($user_id, 'wf_ga_secret_key', $secret_key);
  } // user_options_update


  // enqueue CSS and JS scripts on plugin's admin page
  static function enqueue_scripts() {
    $screen = get_current_screen();

    if (self::is_plugin_page() ||
        $screen->base == 'user' ||
        $screen->base == 'users' ||
        $screen->base == 'profile' ||
        $screen->base == 'user-edit' ||
        get_transient('wf_ga_activation_notice')) {
      $plugin_url = plugin_dir_url(__FILE__);
      wp_enqueue_script('wf-ga', $plugin_url . 'js/wf-ga-admin.js', array('jquery'), WF_GA_VER, true);
      wp_localize_script('jquery', 'wf_ga', array('ajax_nonce' => wp_create_nonce('ajax')));
    } // if
  } // enqueue_scripts


  // set default options
  static function default_settings($force = false) {
    $defaults = array('disable_on_rpc' => '1',
                      'auto_logout' => '15',
                      'secret_login' => self::generate_secret_key(true),
                      'send_qr_users' => '1',
                      'max_login_attempts' => '5',
                      'max_login_attempts_time' => '5',
                      'ban_time' => '120', // 2h
                      'banned_users' => '2', // can't login
                      'ip_whitelist' => '');

    $options = get_option(WF_GA_OPTIONS);

    if ($force || !$options || !$options['banned_users']) {
      update_option(WF_GA_OPTIONS, $defaults);
    }
  } // default_settings


  // sanitize plugin's settings on save
  static function sanitize_settings($values) {
    $old_options = get_option(WF_GA_OPTIONS);

    foreach ($values as $key => $value) {
      switch ($key) {
        case 'disable_on_rpc':
        case 'send_qr_users':
        case 'auto_logout':
        case 'max_login_attempts':
        case 'max_login_attempts_time':
        case 'ban_time':
        case 'banned_users':
          $values[$key] = (int) $value;
        break;
        case 'ip_whitelist':
          $values[$key] = trim($value);
        break;
      } // switch
    } // foreach

    self::check_var_isset($values, array('disable_on_rpc' => 0, 'send_qr_users' => 0));

    if (strlen($values['secret_login']) != 16) {
      $values['secret_login'] = self::generate_secret_key(true);
    }

    return array_merge($old_options, $values);
  } // sanitize_settings


  // check if user is bypassing plugin auth
  static function bypass_auth() {
    $options = get_option(WF_GA_OPTIONS);

    if ((isset($_GET['bypass_auth']) && $_GET['bypass_auth'] == $options['secret_login']) ||
        (isset($_SESSION['bypass_auth']) && $_SESSION['bypass_auth'] === true)) {
      $_SESSION['bypass_auth'] = true;
      return true;
    } else {
      return false;
    }
  } // bypass_auth


  // add authenticator field to login form
  static function login_form() {
    if (self::bypass_auth()) {
      return;
    }

    $field_name = 'wf_ga_otp_' . substr(session_id(), 0, 8);

    echo '<p><label for="' . $field_name . '">' . __('Google Authenticator OTP', 'wf_ga') . '<br />';
    echo '<input type="text" name="' . $field_name . '" id="' . $field_name . '" class="input" value="" size="20" /></label></p>';
  } // login_form


  // remove "remember me" option from login form
  // hack because unfortunately there's no filter for it
  static function login_footer() {
    if (self::bypass_auth()) {
      return;
    }

    echo "\n" . '<script type="text/javascript"> ';
    echo 'try{ ';
    echo 'tmp = document.getElementsByClassName(\'forgetmenot\')[0]; ';
    echo 'tmp.parentNode.removeChild(tmp); ';
    echo '} catch(e){}';
    echo ' </script>' . "\n";
  } // login_footer


  // all settings are saved in one option key
  static function register_settings() {
    register_setting(WF_GA_OPTIONS, WF_GA_OPTIONS, array(__CLASS__, 'sanitize_settings'));
  } // register_settings


  // helper static function for creating dropdowns
  static function create_select_options($options, $selected = null, $output = true) {
    $out = "\n";

    foreach ($options as $tmp) {
      if ($selected == $tmp['val']) {
        $out .= "<option selected=\"selected\" value=\"{$tmp['val']}\">{$tmp['label']}&nbsp;</option>\n";
      } else {
        $out .= "<option value=\"{$tmp['val']}\">{$tmp['label']}&nbsp;</option>\n";
      }
    } // foreach

    if ($output) {
      echo $out;
    } else {
      return $out;
    }
  } // create_select_options


    // helper static function for $_POST checkbox handling
  static function check_var_isset(&$values, $variables) {
    foreach ($variables as $key => $value) {
      if (!isset($values[$key])) {
        $values[$key] = $value;
      }
    }
  } // check_var_isset


  // display warning if WP is outdated
  static function min_version_error_wp() {
    echo '<div class="error"><p>' . sprintf(__('5sec Google Authenticator <b>requires WordPress version 3.6</b> or higher to static function properly. You are using WordPress version %s.', 'wf_ga'), get_bloginfo('version')) . ' Please <a href="' . admin_url('update-core.php') . '">update</a> WordPress.</p></div>';
  } // min_version_error_wp

  // display warning if WP is outdated
  static function activation_notice() {
    echo '<div id="wf_ga_activation_notice" class="updated"><p>' . sprintf(__('<strong>Please read</strong> the <a target="_blank" href="%s">5sec Google Authenticator getting started guide</a>. It is extremly short and up-to-point. It will help you avoid most problems and secure your site in seconds.', 'wf_ga'), plugin_dir_url(__FILE__) . 'documentation/#section-1') . '<i>(<a id="wf_ga_remove_activation_notice" href="#">' . __('hide this message', 'wf_ga') . '</a>)</i></p></div>';
  } // activation notice


  // add authenticator related settings when new user is created
  static function user_register($user_id) {
    update_user_option($user_id, 'wf_ga_secret_key', self::generate_secret_key());
    update_user_option($user_id, 'wf_ga_enable_two_step', '1');
  } // user_register


  // activate plugin
  static function activate() {
    // create tables
    global $wpdb;
    require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

    $table_log = $wpdb->prefix . WF_GA_TABLE_LOG;
    $table_ban = $wpdb->prefix . WF_GA_TABLE_BAN;

    if ($wpdb->get_var("SHOW TABLES LIKE '$table_log'") != $table_log) {
      $sql = "CREATE TABLE `{$table_log}` (
              `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
              `ip` varchar(15) CHARACTER SET latin1 DEFAULT NULL,
              `timestamp` datetime DEFAULT NULL,
              PRIMARY KEY (`id`),
              KEY `ip` (`ip`)
            ) ENGINE=MyISAM DEFAULT CHARSET=utf8;";
      dbDelta($sql);
    }
    if ($wpdb->get_var("SHOW TABLES LIKE '$table_ban'") != $table_ban) {
      $sql = "CREATE TABLE `{$table_ban}` (
              `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
              `ip` varchar(15) CHARACTER SET latin1 DEFAULT NULL,
              `banned_until` datetime DEFAULT NULL,
              PRIMARY KEY (`id`),
              UNIQUE KEY `ip` (`ip`)
            ) ENGINE=MyISAM DEFAULT CHARSET=utf8;";
      dbDelta($sql);
    }

    $users = get_users();
    foreach ($users as $user) {
      update_user_option($user->ID, 'wf_ga_secret_key', self::generate_secret_key());
      update_user_option($user->ID, 'wf_ga_enable_two_step', '1');
    } // foreach user

    self::default_settings(false);
    $options = get_option(WF_GA_OPTIONS);

    // send email to current user
    $current_user = wp_get_current_user();
    $qr_img = self::get_qr_string($current_user->ID);
    $qr_img = $qr_img['img_url'];
    $message = __('5sec Google Authenticator plugin is activated! <b>All users</b> that login from this point forward have to enter their username, password and the OTP (one time password) generated by Google Authenticator app.<br><br> You have received this email as an administrator, and the person who activated the plugin, for two reasons; to immediately setup your access by installing Google Authenticator app and to have access to a secret URL that bypasses OTP usage.<br><br>
<b>Setting up access</b>
<ol>
  <li>Install Google Authenticator app for <a target="_blank" href="https://itunes.apple.com/hr/app/google-authenticator/id388497605?mt=8">iPhone</a>, <a target="_blank" href="https://play.google.com/store/apps/details?id=com.google.android.apps.authenticator2">Android</a> or BlackBerry</li>
  <li>Open the app, click "add an account" (plus sign icon) and choose "scan barcode". Point your phone\'s camera to the QR code on the bottom of this email. This QR is unique for you WP account only! Other users have their own QR codes visible under their profile; or if they register a new account they get it via email.</li>
  <li>You\'re done! Next time you want to login enter the usual username &amp; password combination and the OTP the app generates. Please note that each OTP is valid for only 2 minutes so always use a fresh one.</li>
</ol><br>', 'wf_ga');

$message .= '<b>Secret login URL</b><br>
If you are having problems logging in you can bypass the two step login by using the URL below. Please don\'t share the URL. It can be changed in the plugin\'s settings.<br> Secret URL: <a href="' . site_url() . '/wp-login.php?bypass_auth=' . $options['secret_login'] . '">' . site_url() . '/wp-login.php?bypass_auth=' . $options['secret_login']  . '</a><br><br>QR code for user "' . $current_user->user_login . '"';

    $message .= '<img src="' . $qr_img . '" /><br>';
    $message .= 'If the QR code image is not showing <a href="' . $qr_img . '">open it in browser</a>.<br>';

    wp_mail($current_user->user_email, __('IMPORTANT - 5sec Google Authenticator plugin is activated', 'wf_ga'), $message, 'Content-type: text/html');

    // display notice on plugin usage
    set_transient('wf_ga_activation_notice', true, YEAR_IN_SECONDS);
  } // activate


  // clean-up when deactivated
  static function deactivate() {
    global $wpdb;
    $users = get_users();

    foreach ($users as $user) {
      delete_user_option($user->ID, 'wf_ga_secret_key');
    } // foreach user

    delete_option(WF_GA_OPTIONS);

    // remove tables
    $wpdb->query('DROP TABLE IF EXISTS ' . $wpdb->prefix . WF_GA_TABLE_BAN);
    $wpdb->query('DROP TABLE IF EXISTS ' . $wpdb->prefix . WF_GA_TABLE_LOG);
  } // deactivate
} // wf_ga class


// hook everything up
add_action('init', array('wf_ga', 'init'));

// texdomain has to be loaded earlier
add_action('plugins_loaded', array('wf_ga', 'plugins_loaded'));

// setup environment when activated
register_activation_hook(__FILE__, array('wf_ga', 'activate'));

// when deativated clean up
register_deactivation_hook(__FILE__, array('wf_ga', 'deactivate'));